@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import copy
 import logging
 import math
@@ -9,13 +11,19 @@ from pathlib import Path
 from typing import Literal, cast
 
 from deepmerge import always_merger
-from typing_extensions import TypeAlias, TypedDict, TypeVarTuple
+from typing_extensions import TypeAlias, TypedDict
 
-from ._output import SubmitOutput
+from .. import _env
+from ._util import (
+    Submission,
+    _emit_on_exit_commands,
+    _set_default_envs,
+    _write_run_metadata_commands,
+    _write_submission_meta,
+)
 
 log = logging.getLogger(__name__)
 
-TArgs = TypeVarTuple("TArgs")
 
 _Path: TypeAlias = str | Path | os.PathLike
 MailType: TypeAlias = Literal[
@@ -238,6 +246,14 @@ class SlurmJobKwargs(TypedDict, total=False):
     The setup commands to run before the job.
 
     These commands will be executed prior to everything else in the job script.
+    They will be included in both the worker script (the one that runs inside of srun/jsrun)
+    and not the submission script (the one that is submitted to the job scheduler).
+    """
+
+    submission_script_setup_commands: Sequence[str]
+    """
+    Same as `setup_commands`, but only for the submission script (and not
+    the worker scripts executed by srun/jsrun).
     """
 
     environment: Mapping[str, str]
@@ -259,9 +275,34 @@ class SlurmJobKwargs(TypedDict, total=False):
     The flags to pass to the `srun` command.
     """
 
+    emit_metadata: bool
+    """
+    Whether to emit metadata about the job submission.
+
+    If True (default), the following metadata will be written to a JSON file:
+    - When you submit the job, the job submission information (command, script path, number of jobs, config, and environment) will be written to a file named `submission.json` in the `meta` directory.
+    - When the job starts running, the job ID and environment variables will be written to a file named `run.json` in the `meta` directory.
+    """
+
+    on_exit_script_support: bool
+    """
+    Whether to support running an on-exit script outside of srun.
+
+    This is done by setting the environment variable `NSHRUNNER_EXIT_SCRIPT_DIR` to the path of an initially empty directory.
+    Whenever the script wants something to be done on exit, it should write a bash script to this directory.
+    """
+
+    _exit_script_dir: Path
+    """
+    The directory to write the on-exit scripts to. (Internal use only)
+    """
+
 
 DEFAULT_KWARGS: SlurmJobKwargs = {
-    "name": "ll",
+    "name": "nshrunner",
+    "environment": {
+        _env.SUBMIT_INTERFACE_MODULE: __name__,
+    },
     # "nodes": 1,
     # "time": timedelta(hours=2),
     "timeout_signal": signal.SIGURG,
@@ -269,6 +310,8 @@ DEFAULT_KWARGS: SlurmJobKwargs = {
     "preempt_signal": signal.SIGTERM,
     "open_mode": "append",
     # "requeue": True,
+    "emit_metadata": True,
+    "on_exit_script_support": True,
 }
 
 
@@ -337,6 +380,7 @@ def _write_batch_script_to_file(
     path: Path,
     kwargs: SlurmJobKwargs,
     command: str,
+    env: Mapping[str, str] = {},
     job_array_n_jobs: int | None = None,
 ):
     with path.open("w") as f:
@@ -439,13 +483,41 @@ def _write_batch_script_to_file(
 
         f.write("\n")
 
+        if env:
+            for key, value in env.items():
+                f.write(f"export {key}={value}\n")
+            f.write("\n")
+
+        if (
+            setup_commands := kwargs.get("submission_script_setup_commands")
+        ) is not None:
+            for setup_command in setup_commands:
+                f.write(f"{setup_command}\n")
+            f.write("\n")
+
         if (command_prefix := kwargs.get("command_prefix")) is not None:
             command = " ".join(
                 x_stripped
                 for x in (command_prefix, command)
                 if (x_stripped := x.strip())
             )
-        f.write(f"{command}\n")
+        if not kwargs.get("on_exit_script_support"):
+            f.write(f"{command}\n")
+        else:
+            if (exit_script_dir := kwargs.get("_exit_script_dir")) is None:
+                raise ValueError(
+                    "on_exit_script_support is enabled, but _exit_script_dir is not set. "
+                    "This is a logic error and should not happen."
+                )
+
+            f.write(f"{command} &\n")
+            f.write("wait\n")
+
+            # Add the on-exit script support
+            # Basically, this just emits bash code that iterates
+            # over all files in the exit script directory and runs them
+            # in a subshell.
+            _emit_on_exit_commands(f, exit_script_dir)
 
 
 def update_options(kwargs_in: SlurmJobKwargs, base_dir: Path):
@@ -511,19 +583,38 @@ def update_options(kwargs_in: SlurmJobKwargs, base_dir: Path):
     # Add the command prefix to the kwargs.
     kwargs["command_prefix"] = " ".join(command_parts)
 
-    # Update the environment variables to include the timeout signal
-    if (signal := kwargs.get("timeout_signal")) is not None:
-        kwargs["environment"] = always_merger.merge(
-            kwargs.get("environment", {}),
-            {"NSHRUNNER_TIMEOUT_SIGNAL": signal.name},
+    # Set the default environment variables
+    kwargs["environment"] = _set_default_envs(
+        kwargs.get("environment"),
+        job_index="$SLURM_ARRAY_TASK_ID",
+        local_rank="$SLURM_LOCALID",
+        global_rank="$SLURM_PROCID",
+        world_size="$SLURM_NTASKS",
+        base_dir=base_dir,
+        timeout_signal=kwargs.get("timeout_signal"),
+        preempt_signal=kwargs.get("preempt_signal"),
+    )
+
+    # Emit the setup commands for run metadata
+    if kwargs.get("emit_metadata"):
+        kwargs["setup_commands"] = _write_run_metadata_commands(
+            kwargs.get("setup_commands"),
+            is_worker_script=True,
+        )
+        kwargs["submission_script_setup_commands"] = _write_run_metadata_commands(
+            kwargs.get("submission_script_setup_commands"),
+            is_worker_script=False,
         )
 
-    # Update the environment variables to include the timeout signal
-    if (signal := kwargs.get("preempt_signal")) is not None:
-        kwargs["environment"] = always_merger.merge(
-            kwargs.get("environment", {}),
-            {"NSHRUNNER_PREEMPT_SIGNAL": signal.name},
-        )
+    # If `on_exit_script_support` is enabled, set the environment variable for EXIT_SCRIPT_DIR
+    if kwargs.get("on_exit_script_support"):
+        exit_script_dir = base_dir / "exit_scripts"
+        exit_script_dir.mkdir(exist_ok=True)
+        kwargs["environment"] = {
+            **kwargs.get("environment", {}),
+            _env.EXIT_SCRIPT_DIR: str(exit_script_dir.absolute()),
+        }
+        kwargs["_exit_script_dir"] = exit_script_dir
 
     return kwargs
 
@@ -533,27 +624,36 @@ def to_array_batch_script(
     *,
     script_path: Path,
     num_jobs: int,
-    job_index_variable: str | None = "SLURM_ARRAY_TASK_ID",
     config: SlurmJobKwargs,
-) -> SubmitOutput:
+    env: Mapping[str, str],
+) -> Submission:
     """
     Create the batch script for the job.
     """
     if not isinstance(command, str):
         command = " ".join(command)
 
-    # Update the command to set __NSHRUNNER_JOB_IDX__ to the job index variable (if exists)
-    if job_index_variable:
-        command = command.replace("__NSHRUNNER_JOB_IDX__", job_index_variable)
+    # Write the submission information to a JSON file
+    if config.get("emit_metadata"):
+        _write_submission_meta(
+            script_path.parent,
+            command=command,
+            script_path=script_path,
+            num_jobs=num_jobs,
+            config=config,
+            env=env,
+        )
 
+    # Write the batch script to the file
     _write_batch_script_to_file(
         script_path,
         config,
         command,
+        env=env,
         job_array_n_jobs=num_jobs,
     )
     script_path = script_path.resolve().absolute()
-    return SubmitOutput(
+    return Submission(
         command_parts=["sbatch", f"{script_path}"],
         script_path=script_path,
     )
