@@ -1,16 +1,8 @@
+from __future__ import annotations
+
 from pathlib import Path
 
 
-def _find_git_dir(self):
-    current_dir = Path.cwd()
-    while current_dir != current_dir.parent:
-        git_dir = current_dir / ".git"
-        if git_dir.is_dir():
-            return git_dir
-        current_dir = current_dir.parent
-    return None
-
-
 def _gitignored_dir(path: Path, *, create: bool = True) -> Path:
     if create:
         path.mkdir(exist_ok=True, parents=True)
